package com.gihon.sso.service.impl;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;

import com.fasterxml.jackson.core.type.TypeReference;
import com.gihon.common.util.JacksonJsonUtils;
import com.gihon.common.util.UUIDGenerater;
import com.gihon.sso.entity.vo.LoginUser;
import com.gihon.sso.entity.vo.SsoToken;
import com.gihon.sso.entity.vo.UserInfo;
import com.gihon.sso.service.TokenService;
import com.gihon.sso.service.UserService;

import lombok.extern.slf4j.Slf4j;

//最好TokenStore:有各种实现,类似Cas中的统一认证中心
@Slf4j
@Service
public class TokenServiceImpl implements TokenService {
	
	@Autowired
	private UserService userService;
	@Autowired
	private RedisTemplate<String, String> redisTemplate;
	
	//Redis存储token
	private void saveToken(SsoToken tokenEntity) {
		//TODO 1是否允许多个地点使用同一个Token,2或者每次使用不同的Token。3或者只允许在一个地点登录。
		String userAccount = tokenEntity.getUserInfo().getUserAccount();
		List<SsoToken> tokenList = getTokenByAccount(userAccount);
		if(tokenList==null) {
			tokenList = new ArrayList<SsoToken>();
		}
		tokenList.remove(tokenEntity);
		tokenList.add(tokenEntity);
		redisTemplate.opsForValue().set(
				TOKEN_PRE+tokenEntity.getToken(),
				JacksonJsonUtils.writeObject(tokenEntity),
				TOKEN_EXPIRED+EXIPRED_MORE,
				TimeUnit.MILLISECONDS);
		redisTemplate.opsForValue().set(
				REFRESH_TOKEN_PRE+tokenEntity.getRefreshToken(),
				JacksonJsonUtils.writeObject(tokenEntity),
				REFRESH_TOKEN_EXPIRED+EXIPRED_MORE,
				TimeUnit.MILLISECONDS);
		redisTemplate.opsForValue().set(
				USER_TOKEN_PRE+userAccount,
				JacksonJsonUtils.writeObject(tokenList));
	
	}
	// 移除Redis中token
	private void removeToken(SsoToken tokenEntity) {
		String userAccount = tokenEntity.getUserInfo().getUserAccount();
		List<SsoToken> tokenList = getTokenByAccount(userAccount);
		if(tokenList!=null) {
			tokenList.remove(tokenEntity);
			redisTemplate.opsForValue().set(USER_TOKEN_PRE+userAccount,JacksonJsonUtils.writeObject(tokenList));
		}
		redisTemplate.delete(REFRESH_TOKEN_PRE+tokenEntity.getRefreshToken());
		redisTemplate.delete(TOKEN_PRE+tokenEntity.getToken());
		notifyToken(tokenEntity);
	}
	/**
	 * 通知其他系统Token失效,使用同一个Token的时候有用。需要其他系统重新登录
	 * @param tokenEntity
	 */
	private void notifyToken(SsoToken tokenEntity) {
		return;
	}
	/**
	 * 通知各个系统refreshToken,需要其他系统重新登录
	 * @param tokenEntity
	 * @param tokenEntityNew
	 */
	private void notifyRefreshToken(SsoToken tokenEntity,SsoToken tokenEntityNew) {
		return;
	}
	@Override
	public SsoToken createToken(UserInfo userInfo) {
		if(userInfo==null||!StringUtils.hasText(userInfo.getUserAccount())) {
			return null;
		}
		SsoToken tokenEntity = new SsoToken();
		tokenEntity.setUserInfo(userInfo);
		tokenEntity.setToken(UUIDGenerater.genUUID());
		tokenEntity.setRefreshToken(UUIDGenerater.genUUID());
		long now = System.currentTimeMillis();
		tokenEntity.setRefreshTokenExpire(now+REFRESH_TOKEN_EXPIRED);
		tokenEntity.setTokenExpire(now+TOKEN_EXPIRED);
		saveToken(tokenEntity);
		return tokenEntity;
	}
	
	@Override
	public SsoToken refreshToken(String refreshToken) {
		String tokenValue = redisTemplate.opsForValue().get(REFRESH_TOKEN_PRE+refreshToken);
		
		SsoToken tokenEntity = null;
		
		if(!StringUtils.hasText(tokenValue)) {
			log.warn("refresh_token({})已经失效",refreshToken);
			return tokenEntity;
		}
		tokenEntity = JacksonJsonUtils.readObject(tokenValue, SsoToken.class);
		if(tokenEntity==null) {
			log.warn("refresh_token({})解析失败({})",refreshToken,tokenValue);
			return tokenEntity;
		}
		long now = System.currentTimeMillis();
		if(tokenEntity.getRefreshTokenExpire()<now) {
			log.warn("refresh_token({})已经过期({})",refreshToken,tokenEntity.getRefreshTokenExpire());
			return null;
		}
		removeToken(tokenEntity);
		SsoToken tokenEntityNew = this.createToken(tokenEntity.getUserInfo());
		//TODO 发送通知更新Token
		notifyRefreshToken(tokenEntity,tokenEntityNew);
		return tokenEntityNew;
	}

	@Override
	public UserInfo checkToken(LoginUser loginUser) {
		String tokenValue = redisTemplate.opsForValue().get(TOKEN_PRE+loginUser.getToken());
		SsoToken tokenEntity = null;
		if(StringUtils.hasText(tokenValue)) {
			tokenEntity = JacksonJsonUtils.readObject(tokenValue, SsoToken.class);
			if(tokenEntity!=null) {
				return tokenEntity.getUserInfo();
			}else {
				redisTemplate.delete(TOKEN_PRE+loginUser.getToken());
			}
		}
		return null;
	}

	@Override
	public List<SsoToken> getTokenByAccount(String account) {
		String value = redisTemplate.opsForValue().get(USER_TOKEN_PRE+account);
		List<SsoToken> list = null;
		if(StringUtils.hasText(value)) {
			TypeReference<List<SsoToken>> type = new TypeReference<List<SsoToken>>() {};
			list = JacksonJsonUtils.readObject(value, type);
		}
		return list;
	}

	@Override
	public void clearToken(String token) {
		String tokenValue = redisTemplate.opsForValue().get(TOKEN_PRE+token);
		SsoToken tokenEntity = null;
		if(StringUtils.hasText(tokenValue)) {
			tokenEntity = JacksonJsonUtils.readObject(tokenValue, SsoToken.class);
		}else {
			log.warn("{} token is missing",token);
			return;
		}

		if(tokenEntity!=null) {
			String refreshToken = tokenEntity.getRefreshToken();
			redisTemplate.delete(REFRESH_TOKEN_PRE+refreshToken);
			redisTemplate.delete(TOKEN_PRE+token);
			String userAccount = tokenEntity.getUserInfo().getUserAccount();
			List<SsoToken> tokenList = getTokenByAccount(userAccount);
			if(tokenList!=null) {
				tokenList.remove(tokenEntity);
				redisTemplate.opsForValue().set(USER_TOKEN_PRE+userAccount,JacksonJsonUtils.writeObject(tokenList));
			}
		}
	}

}
